/**
 *
 * Project: Eight Track, File: GuildMusicHandler.java
 *
 * Copyright 2019 Locusworks LLC.
 * All rights reserved. Federal copyright law prohibits unauthorized reproduction by
 * any means and imposes fines up to $25,000 for violation.  No part of this material
 * may be reproduced, transmitted, transcribed, stored in a retrieval system, copied,
 * modified, duplicated, adapted or translated into another program language in any
 * form or by any means, electronic, mechanical, photocopying, recording, or
 * otherwise, without the prior written permission from Locusworks. Locusworks
 * affirms that Eight-Track(R) software and data is subject to United States
 * Government Purpose Rights.  Contact Locusworks, 1313 Lawnview Drive
 * Forney TX 75126, (802) 488-0438, for commercial licensing opportunities.
 *
 * IN NO EVENT SHALL LOCUSWORKS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LOCUSWORKS HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  NO RESPONSIBILITY IS ASSUMED BY
 * LOCUSWORKS FOR ITS USE, OR FOR ANY INFRINGEMENTS OF PATENTS OR OTHER RIGHTS OF
 * THIRD PARTIES RESULTING FROM ITS USE.  LOCUSWORKS SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE AND
 * ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
 * IS". LOCUSWORKS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 */
package net.locusworks.discord.eighttrack.handlers;

import java.awt.Color;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.OffsetDateTime;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.stream.Collectors;

import com.sedmelluq.discord.lavaplayer.player.AudioPlayer;
import com.sedmelluq.discord.lavaplayer.player.DefaultAudioPlayerManager;
import com.sedmelluq.discord.lavaplayer.source.AudioSourceManagers;
import com.sedmelluq.discord.lavaplayer.track.AudioTrack;

import net.dv8tion.jda.api.EmbedBuilder;
import net.dv8tion.jda.api.Permission;
import net.dv8tion.jda.api.entities.ChannelType;
import net.dv8tion.jda.api.entities.GuildChannel;
import net.dv8tion.jda.api.entities.Message.Attachment;
import net.dv8tion.jda.api.entities.MessageEmbed;
import net.dv8tion.jda.api.entities.VoiceChannel;
import net.dv8tion.jda.api.events.message.guild.GuildMessageReceivedEvent;
import net.dv8tion.jda.api.managers.AudioManager;
import net.locusworks.discord.eighttrack.scheduler.TrackScheduler;
import net.locusworks.logger.ApplicationLogger;
import net.locusworks.logger.ApplicationLoggerFactory;

public class GuildMusicHandler {

  private Path musicDir;
  private TrackScheduler ts;
  private DefaultAudioPlayerManager apm;
  private ApplicationLogger logger;
  private AudioPlayer player;
  private AtomicBoolean playing;
  private Long voiceChannelId;
  
  private OffsetDateTime lastPlayed;
  
  private Consumer<Long> callback;
  private Mp3UploadHandler uploadHandler;
  
  public GuildMusicHandler(Path musicDir, Mp3UploadHandler uploadHandler) throws IOException {
    this.logger = ApplicationLoggerFactory.getLogger(GuildMusicHandler.class);
    this.playing = new AtomicBoolean(false);
    this.musicDir = musicDir;
    this.lastPlayed = OffsetDateTime.now();
    this.uploadHandler = uploadHandler;
    this.apm = new DefaultAudioPlayerManager();
    
    AudioSourceManagers.registerLocalSource(apm);
    this.player = apm.createPlayer();
    
    this.ts = new TrackScheduler();
    player.addListener(ts);
  }
  
  public OffsetDateTime getLastPlayed() {
    return lastPlayed;
  }
  
  public void accept(Consumer<Long> callback) {
    this.callback = callback;
  }
  
  public void isPlaying(boolean playing) {
    this.playing.set(playing);
  }
  
  public void indexFiles()  throws IOException {
    List<Path> songs = Files.walk(musicDir)
        .filter(p-> p.getFileName().toString().endsWith(".mp3"))
        .collect(Collectors.toList());

    logger.info("Adding %d songs from directory %s", songs.size(), musicDir);

    for (Path song : songs) {
      logger.debug("Loading song: %s", song);
      try {
        apm.loadItem(song.toAbsolutePath().toString(), ts);
      } catch (IllegalStateException ex) {
        logger.warn("Unable to load song :%s -> %s", song.toAbsolutePath().toString(), ex.getMessage());
      }
    }
  }
  
  public void index(GuildMessageReceivedEvent event) {
    event.getChannel().sendMessage(String.format("<@%s>, Please wait as i reindex the music files", event.getAuthor().getId())).queue();
    try {
      indexFiles();
    } catch (IOException e) {
      event.getChannel().sendMessage(String.format("<@%s>, Unable to reindex the music files: %s", event.getAuthor().getId(), e.getMessage())).queue();
      return;
    }

    event.getChannel().sendMessage(String.format("<@%s>, Reindexing complete", event.getAuthor().getId())).queue();
  }

  public void whatsNext(GuildMessageReceivedEvent event) {
    AudioTrack track = ts.peek();
    MessageEmbed embed = new EmbedBuilder()
        .setAuthor(event.getMember().getEffectiveName(), null, event.getAuthor().getAvatarUrl())
        .setTitle("Next Up:")
        .setDescription(String.format("**%s** by __%s__", track.getInfo().title, track.getInfo().author))
        .setTimestamp(OffsetDateTime.now())
        .build();
    event.getChannel().sendMessage(embed).queue();
  }

  public void repeat(GuildMessageReceivedEvent event) {
    next(event);
    ts.setFinishedCallback((ater) -> {
      if (!playing.get()) return;
      next(event);
    });
  }

  public void next(GuildMessageReceivedEvent event) {
    
    GuildChannel gc = event.getGuild().getGuildChannelById(ChannelType.VOICE, voiceChannelId);
    if (gc != null && gc.getMembers().size() == 1) {
      event.getChannel().sendMessage("Going silent since no one is currently listening to the channel").queue();
      event.getGuild().getAudioManager().closeAudioConnection();
      if (callback != null) callback.accept(event.getGuild().getIdLong());
      return;
    }
    
    playing.set(false);
    stop(event);
    play(event);
    playing.set(true);
  }

  public void stop(GuildMessageReceivedEvent event) {
    player.stopTrack();
  }

  public void play(GuildMessageReceivedEvent event) {
    VoiceChannel vc = event.getMember().getVoiceState().getChannel();
    if (vc == null) {
      event.getChannel().sendMessage(String.format("<@%s> you are not in a voice channel to play music", event.getMember().getId())).queue();
      return;
    }
    
    if (!event.getGuild().getSelfMember().hasPermission(vc, Permission.VOICE_CONNECT, Permission.VOICE_SPEAK)) {
      event.getChannel().sendMessage(String.format("<@%s>, I cannot play my music in channel %s as I dont have permission to", 
          event.getMember().getId(), vc.getName())).queue();
      return;
    }
    
    voiceChannelId = vc.getIdLong();

    AudioManager manager = event.getGuild().getAudioManager();
    manager.openAudioConnection(vc);

    manager.setSendingHandler(new EightTrackAudioSendHandler(player));

    if (ts.hasTracks()) {
      lastPlayed = OffsetDateTime.now(); 
      
      AudioTrack track = ts.getNextTrack();
      
      MessageEmbed embed = new EmbedBuilder()
          .setAuthor(event.getMember().getEffectiveName(), null, event.getAuthor().getAvatarUrl())
          .setTitle("Now Playing:")
          .setDescription(String.format("**%s** by __%s__", track.getInfo().title, track.getInfo().author))
          .setTimestamp(OffsetDateTime.now())
          .build();

      event.getChannel().sendMessage(embed).queue();
      player.playTrack(track);
      ts.trackLoaded(track);
    }
  }

  public void upload(GuildMessageReceivedEvent event) {
    if (!event.getMessage().getMember().hasPermission(Permission.ADMINISTRATOR, Permission.MANAGE_SERVER)) {
      event.getChannel().sendMessage(String.format("Sorry <@%s> i can't do that. *psst: you have to have manage server, or administrator role*", event.getMember().getIdLong())).queue();
      return;
    }
    
    for(Attachment attachment : event.getMessage().getAttachments()) {
      attachment.retrieveInputStream().thenAccept((in) -> {
        try {
          Mp3UploadResults res = uploadHandler.parse(in);
          
          MessageEmbed embed = new EmbedBuilder()
              .setColor(Color.GREEN)
              .setTitle("Upload Results")
              .setDescription(res.toString())
              .setTimestamp(OffsetDateTime.now())
              .setFooter(event.getGuild().getSelfMember().getEffectiveName(), event.getGuild().getSelfMember().getUser().getAvatarUrl())
              .build();
          
          event.getChannel().sendMessage(embed).queue();
        } catch (Exception ex) {
          logger.error("Unable to get file information: %s", ex.getMessage());
          logger.error(ex);
        } finally {
          event.getMessage().delete().queue();
        }
        
      }).exceptionally((err) ->{
        return null;
      });
    }
    
  }
}
